#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_eigen.h>
#include "logLike.h"
#include "loadData.h"

typedef struct {
    gsl_vector *param;
    int var, var2;
} dparam;
dparam dfparam;
gsl_vector *tparam;

double singleNLikelihood(double x, void *params){
    gsl_vector *param = ((dparam*)params)->param;
    int ind = ((dparam*)params)->var;
    
    double old = gsl_vector_get(param, ind);
    gsl_vector_set(param, ind, x);

    double val = -logLikelihood(param);

    gsl_vector_set(param, ind, old);
    return val;
}

double coupleNLikelihood(double x, double y, void *params){
    gsl_vector *param = ((dparam*)params)->param;
    int ind = ((dparam*)params)->var;
    int ind2 = ((dparam*)params)->var2;
    
    double old1 = gsl_vector_get(param, ind);
    double old2 = gsl_vector_get(param, ind2);
    gsl_vector_set(param, ind, x);
    gsl_vector_set(param, ind2, y);

    double val = -logLikelihood(param);

    gsl_vector_set(param, ind, old1);
    gsl_vector_set(param, ind2, old2);
    return val;
}

void localHessian(gsl_vector *x, gsl_vector *dx, double *deriv, double *hessian){

#define fx(f,x,dx) (-f(x+2*dx) + 8*f(x+1*dx) - 8*f(x-1*dx) + f(x-2*dx))/(12*dx)

#define fxx(f,x,dx) (-f(x+2*dx) + 16*f(x+1*dx) - 30*f(x) \
                                + 16*f(x-1*dx) - f(x-2*dx))/(12*dx*dx)

#define ffx(f,x,y,dx) (-f(x+2*dx,y) + 8*f(x+1*dx,y) \
                    - 8*f(x-1*dx,y) + f(x-2*dx,y))/(12*dx)
#define fxy(f,x,y,dx,dy) (-ffx(f,x,y+2*dy,dx) + 8*ffx(f,x,y+1*dy,dx) \
                    - 8*ffx(f,x,y-1*dy,dx) + ffx(f,x,y-2*dy,dx))/(12*dy)

#define single(x) singleNLikelihood(x, &dfparam)
#define couple(x,y) coupleNLikelihood(x, y, &dfparam)

    int i,j;
    gsl_vector_memcpy(dfparam.param, x);
    for(i = 0; i < nparam; i++){
        dfparam.var = i;
        deriv[i] = fx(single, gsl_vector_get(x, i), gsl_vector_get(dx, i));
    }
    
    for(i = 0; i < nparam; i++){
        dfparam.var = i;
        for(j = 0; j < nparam; j++){
            dfparam.var2 = j;
            if(i == j){
                hessian[j+nparam*i] = fxx(single, gsl_vector_get(x, i), gsl_vector_get(dx, i));
            }
            else{
                hessian[j+nparam*i] = fxy(couple, gsl_vector_get( x, i), gsl_vector_get( x, j), 
                                             gsl_vector_get(dx, i), gsl_vector_get(dx, j));
            }
        }
    }
}

int main (int argc, char *argv[]) {
    int i,j;

    if(argc != 2 + nparam){
        printf("Usage: ./a.out datafile p1 p2 ... (%d params: ", nparam);
        printLikelihoodVars();
        printf(")\n");
        exit(1);
    }
    
    int nx;
    double *x, *background;
    int nEn;
    double *energies, *seqbgp;
    readData(argv[1], &nx, &x, &background, &nEn, &energies, &seqbgp);
    char *outfname = argv[2];

    init(nx, x, background, nEn, energies, seqbgp);

    tparam = gsl_vector_alloc(nparam);
    
    gsl_vector *param, *dparam;
    param = gsl_vector_alloc(nparam);
    dparam = gsl_vector_alloc(nparam);
    for(i = 0; i < nparam; i++){
        gsl_vector_set(param, i, atof(argv[i+2]));
        gsl_vector_set(dparam, i, 0.01);
    }
    mapParam(param);
    dfparam.param = gsl_vector_alloc(nparam);

    double ll = -logLikelihood(param);

    double hessian[nparam*nparam];
    double deriv[nparam];

    localHessian(param,dparam, deriv, hessian);

    printLikelihoodVars();
    printf(":\n");
    unmapParam(param);
    for(i = 0; i < nparam; i++){
        printf("%g  ", gsl_vector_get(param, i));
    }
    printf("(%g)\n", ll);

    printf("\nDerivs:\n[");
    for(i = 0; i < nparam; i++){
        printf("%g, ", deriv[i]);
    }
    printf("]\n\n");

    printf("Hessian:\n");
    printf("[");
    for(i = 0; i < nparam; i++){
        printf(" [");
        for(j = 0; j < nparam; j++){
            printf("%g, ", hessian[j+nparam*i]);
        }
        printf("],\n");
    }
    printf("]\n\n");


    gsl_matrix_view hm = gsl_matrix_view_array(hessian, nparam, nparam);

    gsl_vector *eval = gsl_vector_alloc(nparam);
    gsl_matrix *evec = gsl_matrix_alloc(nparam, nparam);

    gsl_eigen_symmv_workspace *w = gsl_eigen_symmv_alloc(nparam);
    gsl_eigen_symmv(&hm.matrix, eval, evec, w);
    gsl_eigen_symmv_free(w);

    gsl_eigen_symmv_sort(eval, evec, GSL_EIGEN_SORT_ABS_ASC);

    for (i = 0; i < nparam; i++) {
        double eval_i = gsl_vector_get (eval, i);
        gsl_vector_view evec_i = gsl_matrix_column (evec, i);

        printf ("% 9f   ", eval_i);
        for(j = 0; j < nparam; j++){
            printf(" % 8.3f", gsl_vector_get(&evec_i.vector, j));
        }
        printf("\n");
    }
    printf("\n");

    gsl_vector_free (eval);
    gsl_matrix_free (evec);

    return 0;
}

