#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_eigen.h>
#include "logLike.h"
#include "loadData.h"

typedef struct {
    gsl_vector *param;
    int var, var2;
} dparam;
dparam dfparam;
gsl_vector *tparam;

double singleNLikelihood(double x, void *params){
    gsl_vector *param = ((dparam*)params)->param;
    int ind = ((dparam*)params)->var;
    
    double old = gsl_vector_get(param, ind);
    gsl_vector_set(param, ind, x);

    double val = -logLikelihood(param);

    gsl_vector_set(param, ind, old);
    return val;
}

double coupleNLikelihood(double x, double y, void *params){
    gsl_vector *param = ((dparam*)params)->param;
    int ind = ((dparam*)params)->var;
    int ind2 = ((dparam*)params)->var2;
    
    double old1 = gsl_vector_get(param, ind);
    double old2 = gsl_vector_get(param, ind2);
    gsl_vector_set(param, ind, x);
    gsl_vector_set(param, ind2, y);

    double val = -logLikelihood(param);

    gsl_vector_set(param, ind, old1);
    gsl_vector_set(param, ind2, old2);
    return val;
}


void nLogLikedf(const gsl_vector *v, void *params, gsl_vector *g){
    gsl_vector_memcpy(dfparam.param, v);
    gsl_function F;
    F.function = &singleNLikelihood;
    F.params = &dfparam;
    double abserr;
    
    int i;
    for(i = 0; i < nparam; i++){
        double d;
        dfparam.var = i;
        gsl_deriv_central(&F, gsl_vector_get(v, i), 1e-8, &d, &abserr);
        gsl_vector_set(g, i, d);
    }
}

double nLogLikef(const gsl_vector *v, void *params){
    return -logLikelihood(v);
}

void nLogLikefdf(const gsl_vector *v, void *params, double *f, gsl_vector *g){
    nLogLikedf(v,params,g);
    *f = nLogLikef(v,params);
}

double minimizeSimple(gsl_vector *param){
    int i,j,k,n;
    int indj, dirj, indk, dirk;
    double lastmin = INFINITY;
    
    double dx = 32*0.01;

    for(n = 0; n < 10; n++){
        double min = INFINITY;
        int mindirj, mindirk;

        for(j = 0; j < nparam*4; j++){
            indj = j/4;
            dirj = (j%4)-2 + ((j%4)>=2? 1: 0);
            double oldj = gsl_vector_get(param, indj);

            gsl_vector_set(param, indj, gsl_vector_get(param, indj) + dirj*dx);

            for(k = 0; k < nparam*4; k++){
                indk = k/4;
                dirk = (k%4)-2 + ((k%4)>=2? 1: 0);
                double oldk = gsl_vector_get(param, indk);

                gsl_vector_set(param, indk, gsl_vector_get(param, indk) + dirk*dx);
    
                double val = -logLikelihood(param);

                if(val < min){
                    mindirj = j;
                    mindirk = k;
                    min = val;
                }
                gsl_vector_set(param, indk, oldk);
            }
            gsl_vector_set(param, indj, oldj);
        }

        if(min == INFINITY || fabs(lastmin - min) < 1e-10){
            if(dx <= 0.01){
                break;
            }
            dx = dx/2;
            continue;
        }

        indj = mindirj/4;
        dirj = (mindirj%4)-2 + ((mindirj%4)>=2? 1: 0);
        indk = mindirk/4;
        dirk = (mindirk%4)-2 + ((mindirk%4)>=2? 1: 0);
        gsl_vector_set(param, indj, gsl_vector_get(param, indj) + dirj*dx);
        gsl_vector_set(param, indk, gsl_vector_get(param, indk) + dirk*dx);
        lastmin = -logLikelihood(param);

        gsl_vector_memcpy(tparam, param);
        unmapParam(tparam);
        for(i = 0; i < nparam; i++){
            fprintf(stderr, "%g  ", gsl_vector_get(tparam, i));
        }
        fprintf(stderr, "\nSS %d %g   %.15g %.15g\n", n, dx, -logLikelihood(param), lastmin - min);
    }

    return -logLikelihood(param);
}

double minimize(gsl_vector *x){
    int i;
    size_t iter = 0;
    double size;
    int status;

    const gsl_multimin_fminimizer_type *T;
    gsl_multimin_fminimizer *s;

    gsl_multimin_function minfunc;
    minfunc.n = x->size;
    minfunc.f = &nLogLikef;

    T = gsl_multimin_fminimizer_nmsimplex2;
    s = gsl_multimin_fminimizer_alloc(T, x->size);

    gsl_vector *dx;
    dx = gsl_vector_alloc(nparam);
    gsl_vector_set_all(dx, 0.0001);


    gsl_multimin_fminimizer_set(s, &minfunc, x, dx);
    do{
        iter++;
        status = gsl_multimin_fminimizer_iterate (s);

        if(status){
            break;
        }

        gsl_vector_memcpy(tparam, s->x);
        unmapParam(tparam);
        for(i = 0; i < nparam; i++){
            fprintf(stderr, "%g  ", gsl_vector_get(tparam, i));
        }
        fprintf(stderr, "STE2P %g\n", s->fval);

        size = gsl_multimin_fminimizer_size (s);
        status = gsl_multimin_test_size (size, 1e-8);
    }
    while (status == GSL_CONTINUE && iter < 100000);
    
    gsl_vector_memcpy(x, s->x);

    gsl_multimin_fminimizer_free (s);
    return s->fval;
}

double randu(double low, double high, gsl_rng *r){
    return gsl_rng_uniform(r)*(high-low) + low;
}

int main (int argc, char *argv[]) {
    int i,j;

    if(argc != 3 + nparam){
        printf("Usage: ./a.out datafile p1 p2 ... (%d params: ", nparam);
        printLikelihoodVars();
        printf(")\n");
        exit(1);
    }
    
    int nx;
    double *x, *background;
    int nEn;
    double *energies, *seqbgp;
    readData(argv[1], &nx, &x, &background, &nEn, &energies, &seqbgp);
    char *outfname = argv[2];

    init(nx, x, background, nEn, energies, seqbgp);

    tparam = gsl_vector_alloc(nparam);
    
    gsl_vector *param, *dparam;
    param = gsl_vector_alloc(nparam);
    dparam = gsl_vector_alloc(nparam);
    for(i = 0; i < nparam; i++){
        gsl_vector_set(param, i, atof(argv[i+3]));
        gsl_vector_set(dparam, i, 0.01);
    }
    mapParam(param);
    dfparam.param = gsl_vector_alloc(nparam);

    fprintf(stderr, "Initial LL: %g\n", nLogLikef(param, NULL));

    double ll = minimize(param);
    //ll = minimizeSimple(param);

    double hessian[nparam*nparam];
    double deriv[nparam];

    FILE *outf = fopen(outfname, "wb");

    printLikelihoodVars();
    printf(":\n");
    unmapParam(param);
    for(i = 0; i < nparam; i++){
        double p = gsl_vector_get(param, i);
        printf("%g  ", p);
        fwrite(&p, sizeof(double), 1, outf);
    }
    printf("(%g)\n", ll);
    fwrite(&ll, sizeof(double), 1, outf);
    fclose(outf);


    printf("\nDerivs:\n[");
    for(i = 0; i < nparam; i++){
        printf("%g, ", deriv[i]);
    }
    printf("]\n\n");

    return 0;
}

