#!/usr/bin/env python2
from __future__ import with_statement
from scipy import *
import sys, types
import pylab
import itertools

class Opener:
    def __init__(self, fileobj, rw="rt"):
        self.fileobj = fileobj
        self.rw = rw
        self.f = None

    def __enter__(self):
        if isinstance(self.fileobj, types.StringType):
            self.f = open(self.fileobj, self.rw)
            return self.f
        elif hasattr(self.fileobj, 'read') or hasattr(self.fileobj, 'write'):
            if self.rw != self.fileobj.mode: #error? XXX
                raise Exception("File is already open ({}), but in wrong mode ({})".format(self.fileobj.mode, self.rw))
            return self.fileobj

    def __exit__(self, e, fileobj, t):
        if self.f != None:
            self.f.close()
        return False

def getHeader(fn):
    with Opener(fn, 'rt') as f:
        header = list(itertools.takewhile(lambda s: s.startswith('#'), f))
    return header

def parseHeader(hd):
    etable = []
    tree = None
    info = []
    param = None
    for line in hd:
        if line.startswith('#ETABLE '):
            etable.append(line)
        elif line.startswith('#TREE '):
            tree = line[len('#TREE '):]
        elif line.startswith('#PARAM '):
            terms = [kv.split(':') for kv in line[len('#PARAM '):].split(',')]
            param = dict([(k.strip(), eval(v)) for k,v in terms])
        elif line.startswith('#INFO '):
            info.append(line)
    
    etable = array([[float(x) for x in line[len('#ETABLE '):].split()] 
                                   for line in etable])
    
    return info, param, etable, tree

class SeqToNum:
    def __init__(self, names):
        self.nucNums = -ones(256, int)
        self.nucNums[frombuffer(names, uint8)] = arange(len(names))
    
    def __call__(self, seq):
        return self.nucNums[frombuffer(seq, uint8)]

def loadSites(fn, names=None): #optimized for fast loading, assumes ASCII
    if isinstance(fn, types.StringType):
        f = open(fn, 'rt')
    else:
        f = fn

    l = f.next()
    header = []
    while l.startswith('#'):
        header.append(l)
        l = f.next()
    seqs = [l.strip()] + [l.strip() for l in f]
    if seqs[-1] == '':
        del seqs[-1]
    
    if isinstance(fn, types.StringType):
        f.close()
    
    try:
        info, param, etable, tree = parseHeader(header)
    except Exception as e:
        print e
        info, param, etable, tree = None,None,None,None
    if names == None:
        if param == None or 'bases' not in param:
            raise Exception("Could not determine names of bases")
        names = param['bases']

    nSeqs, seqLen = len(seqs), len(seqs[0])
    if any([len(s) != seqLen for s in seqs]): #check that file is OK
        raise Exception("Error: Sequences have different lengths")

    nucNums = -ones(256, int) #nucNums is a map from ascii to base number
    nucNums[frombuffer(names, uint8)] = arange(len(names))
    
    bases = frombuffer("".join(seqs), uint8).reshape((nSeqs, seqLen))
    seqTable = nucNums[bases]

    badBases = seqTable < 0
    if any(badBases):
        raise Exception("Invalid residue(s): {0}".format(" ".join(chr(i) for i in set(bases[badBases]))))

    return seqTable, (names,info,param,etable,tree)

def loadSitesGen(fn, names):
    nucNums = -ones(256, int) #nucNums is a map from ascii to base number
    nucNums[frombuffer(names, uint8)] = arange(len(names))

    data = []
    with Opener(fn, 'rt') as f:
        for line in f:
            if line.startswith('#'):
                continue
            
            yield nucNums[frombuffer(line[:-1], uint8)]

def loadSitesGradual(fn, names, mapper=None):
    #for use when memory is an issue
    if mapper == None:
        mapper = lambda x: x

    nucNums = -ones(256, int) #nucNums is a map from ascii to base number
    nucNums[frombuffer(names, uint8)] = arange(len(names))
    
    data = []
    with Opener(fn, 'rt') as f:
        for line in f:
            if line.startswith('#'):
                continue
            
            seq = nucNums[frombuffer(line[:-1], uint8)]
            data.append(mapper(seq))

    return array(data)

def writeSites(fn, seqs, names, info=None, param=None, etable=None, etree=None):
    bs = []
    with Opener(fn, 'w') as f:
        if param != None:
            pstr = ",".join("{0}: {1}".format(k,repr(v)) for k,v in param.iteritems())
            f.write('#PARAM {0}\n'.format(pstr))
        if info != None:
            f.write('#INFO {0}\n'.format(info))
        if etable != None:
            for r in etable:
                f.write('#ETABLE {0}\n'.format(" ".join(str(c) for c in r)))
        if etree != None:
            f.write('#INFO {0}\n'.format(etree))

        for s in seqs:
            f.write("".join(names[i] for i in s))
            f.write("\n")

def energyFunc(em):
    if isinstance(em, basestring):
        em = loadtxt(em)
    
    seqLen = em.shape[0]
    dummyPP = arange(seqLen)
    def getEnergy(seq):
        return sum(em[dummyPP, seq])
    return getEnergy

def getEnergies(seq, em):
    if isinstance(em, basestring):
        em = loadtxt(em)
    
    nSeq, seqLen = seq.shape
    dummyPP = arange(seqLen, dtype=int).reshape(1,seqLen).repeat(nSeq,0)
    return sum(em[dummyPP, seq], 1)

def getCounts(seqs, nBases):
    nSeq, seqLen = seqs.shape
    bins = arange(nBases+1, dtype='int')
    counts = zeros((seqLen, nBases), dtype='int')
    for i in range(seqLen):
        counts[i,:] = histogram(seqs[:,i], bins)[0]
    return counts # index as [pos, res]

def getFreqs(seq, nBases):
    return getCounts(seq, nBases).astype('float')/seq.shape[0]

import matplotlib.cm
from matplotlib.patches import PathPatch
from matplotlib.text import TextPath
from matplotlib.transforms import Affine2D
from matplotlib.font_manager import FontProperties
def showFreqs(f, bases="ACGT", nSeqs=inf, fig=None):
    #according to logo standard
    seqLen, nBases = f.shape
    if nBases != len(bases):
        raise Exception("Need %d bases" % len(bases))

    if nBases == 4:
        colors = ["#00cc00", "#0000cc", "#ffb300", "#cc0000"]
    else:
        colors = matplotlib.cm.jet(arange(1,nBases+1, dtype=float)/(nBases+1))

    fontprop = FontProperties(family="monospace", stretch="condensed", 
                              weight="bold", size="medium")
    
    #see the scheider papaers on logos (n2Ua and m3WS) for description
    #here we use the 'approximate' method, good for nseq > 50.
    ecorr = (nBases-1)/(2*log(2)*nSeqs)
    f = f/sum(f,1)[:,newaxis]
    old_settings = seterr(divide='ignore', invalid='ignore')
    R = log2(nBases) - (-sum(nan_to_num(f*log2(f)),1)+ecorr)
    seterr(**old_settings)
    heights = f*R[:,newaxis]
    
    if fig == None:
        fig = pylab.figure(figsize=(8,3))
    for n,h in enumerate(heights):
        ord = argsort(h)
        s = 0
        for i in range(nBases):
            height = h[ord[i]]
            text = TextPath((0,0),bases[ord[i]],size=1, prop=fontprop)
            bbox = text.get_extents()
            tx0,ty0 = bbox.min
            tw,th = bbox.width, bbox.height
            trans = Affine2D.identity() \
                            .translate(-tx0,-ty0) \
                            .scale(0.9/tw,height/th) \
                            .translate(n+0.5,s)
            t = trans.transform_path(text)
            pylab.gca().add_patch(PathPatch(t, fc=colors[ord[i]], ec='none'))
            s += height

    pylab.xticks(arange(seqLen)+1)
    pylab.ylabel('Bits')
    pylab.xlim(0.5,seqLen+0.5)
    #pylab.ylim(0,max(sum(heights,1)))
    pylab.ylim(0,log2(nBases))
    return fig

def showEnergies(em, bases="ACGT", nSeqs=inf, fig=None):
    #according to logo standard
    seqLen, nBases = em.shape
    if nBases != len(bases):
        raise Exception("Need %d bases" % len(bases))

    if nBases == 4:
        colors = ["#00cc00", "#0000cc", "#ffb300", "#cc0000"]
    else:
        colors = matplotlib.cm.jet(arange(1,nBases+1, dtype=float)/(nBases+1))

    fontprop = FontProperties(family="monospace", stretch="condensed", 
                              weight="bold", size="medium")
    em = em-mean(em,1)[:,newaxis]
    sem = argsort(em,0)
    
    if fig == None:
        fig = pylab.figure(figsize=(8,3))
    
    def addletter(let, x, y, height, color, alpha=1):
            text = TextPath((0,0),let,size=1, prop=fontprop)
            bbox = text.get_extents()
            tx0,ty0 = bbox.min
            tw,th = bbox.width, bbox.height
            trans = Affine2D.identity() \
                            .translate(-tx0,-ty0) \
                            .scale(0.9/tw,height/th) \
                            .translate(x,y)
            t = trans.transform_path(text)
            pylab.gca().add_patch(PathPatch(t, fc=color, ec='none',alpha=alpha))

    for n,row in enumerate(em):
        pos = arange(nBases)[row>0][argsort(row[row>0])]
        neg = arange(nBases)[row<0][argsort(row[row<0])[::-1]]
            
        hpos = 0
        for i in pos:
            addletter(bases[i], n+0.5, hpos, row[i], colors[i])
            hpos += row[i]
        hpos = 0
        for i in neg:
            addletter(bases[i], n+0.5, hpos+row[i], -row[i], colors[i],alpha=0.5)
            hpos += row[i]
    
    pylab.hlines(0, 0, seqLen+0.5, colors='k')
    pylab.xticks(arange(seqLen)+1)
    #pylab.ylabel('Bits')
    pylab.xlim(0.5,seqLen+0.5)
    #pylab.ylim(0,max(sum(heights,1)))
    #pylab.ylim(0,log2(nBases))
    return fig

def main():
    if len(sys.argv) == 1:
        print "Usage: load.py seq bases [freq|counts|energy em]"
        return

    bases = sys.argv[2]
    seq = loadSites(sys.argv[1], bases)[0]

    if sys.argv[3] == 'freq':
        f = getFreqs(seq, len(bases))
        savetxt(sys.stdout, f)
    elif sys.argv[3] == 'counts':
        c = getCounts(seq, len(bases))
        savetxt(sys.stdout, c)
    elif sys.argv[3] == 'energy':
        em = loadtxt(sys.argv[4])
        print "\n".join(str(e) for e in getEnergies(seq, em))

if __name__ == '__main__':
    main()



#unoptimized but clearer version of loadSites:
#def loadSites(fn, names):
#    with Opener(fn, 'rt') as f:
#        dat = f.read()
#        if dat[0] == '#': #skip comment in first line if present
#            dat = dat[dat.index('\n'):]
#        seqs = dat.split()
#    
#    nucNums = dict(zip(names, range(len(names))))
#    bs = []
#    for s in seqs:
#        bs.append(array([nucNums[n] for n in s]))
#    return array(bs)
